# -*- coding:japanese-shift-jis-dos; tab-width:4; -*-
#
# JargonXml.pm
# $Id: JargonXml.pm 2.0 2004/08/24 21:37:41 satomii Exp $
#
# Copyright (C) 2004 Satomi I.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#

package JargonXml;
use strict;
use IO::Handle;
use XML::LibXML;

# nodeType values.
use constant NODE_ELEMENT			=> 1;
use constant NODE_ATTRIBUTE			=> 2;
use constant NODE_TEXT				=> 3;
use constant NODE_CDATA				=> 4;
use constant NODE_ENTITY_REF		=> 5;
use constant NODE_ENTITY			=> 6;
use constant NODE_PI				=> 7;
use constant NODE_COMMENT			=> 8;
use constant NODE_DOCUMENT			=> 9;
use constant NODE_DOCUMENT_TYPE		=> 10;
use constant NODE_DOCUMENT_FLAG		=> 11;
use constant NODE_DTD				=> 14;
use constant NODE_ELEMENT_DECL		=> 15;
use constant NODE_ATTRIBUTE_DECL	=> 16;
use constant NODE_ENTITY_DECL		=> 17;

use vars qw(@ISA @EXPORT @EXPORT_OK %ENTITY);

@ISA = qw(XML::LibXML);
@EXPORT = qw(
	NODE_ELEMENT NODE_ATTRIBUTE NODE_TEXT NODE_CDATA NODE_ENTITY_REF
	NODE_ENTITY NODE_PI NODE_COMMENT NODE_DOCUMENT NODE_DOCUMENT_TYPE
	NODE_DOCUMENT_FLAG NODE_DTD NODE_ELEMENT_DECL NODE_ATTRIBUTE
	NODE_ENTITY_DECL
);
@EXPORT_OK = qw(%ENTITY);

%ENTITY = (
	0x0097	=> "\x81\x5c",					# &mdash; or &#x2012;
	0x00a3	=> '',						# &pound;
	0x00ad	=> '',							# &shy;, soft hyphen, ignore.
	0x00ae	=> '<userchar name="f00ae"/>',	# &reg;
	0x00b1	=> '}',						# &plusmn;
	0x00b4	=> 'L',						# &acute;
	0x00b5	=> '<userchar name="h00b5"/>',	# &micro;
	0x00b7	=> 'E',						# &middot;
	0x00c9	=> '<userchar name="h00c9"/>',	# &Eacute;
	0x00d7	=> '<userchar name="h00d7">~</userchar>',	# &times;
	0x00d8	=> '<userchar name="h00d8"/>',	# &Oslash;
	0x00df	=> '<userchar name="h00df"/>',	# &szlig;
	0x00e0	=> '<userchar name="h00e0"/>',	# &agrave;
	0x00e1	=> '<userchar name="h00e1"/>',	# &aacute;
	0x00e4	=> '<userchar name="h00e4"/>',	# &auml;
	0x00e6	=> '<userchar name="f00e6"/>',	# &aelig;
	0x00e8	=> '<userchar name="h00e8"/>',	# &egrave;
	0x00e9	=> '<userchar name="h00e9"/>',	# &eacute;
	0x00ea	=> '<userchar name="h00ea"/>',	# &ecirc;
	0x00ef	=> '<userchar name="h00ef"/>',	# &iuml;
	0x00f6	=> '<userchar name="h00f6"/>',	# &ouml;
	0x00fc	=> '<userchar name="h00fc"/>',	# &uuml;
	0x0259	=> '<userchar name="h0259"/>',	# &schwa;
	0x039b	=> '',						# &Lambda;
	0x03b1	=> '',						# &alpha;
	0x03b9	=> '',						# &iota;
	0x03ba	=> '',						# &kappa;
	0x03bb	=> '',						# &lambda;
	0x03bc	=> '',						# &mu;
	0x03bd	=> '',						# &nu;
	0x03bf	=> '',						# &omicron;
	0x03c0	=> '',						# &pi;
	0x03c1	=> '',						# &rho;
	0x2018	=> 'e',						# &lsquo;
	0x2019	=> 'f',						# &rsquo;
	0x201c	=> 'g',						# &ldquo;
	0x201d	=> 'h',						# &rdquo;
	0x2190	=> '',
	0x2192	=> '',
	0x21d4	=> '',
	0x2205	=> '<userchar name="h2205"/>',	# &empty;
	0x2212	=> '-',							# &minus;
	0x223c	=> '`',						# &sim;
	0x2295	=> '<userchar name="f2295"/>',	# &oplus;
	0x2297	=> '<userchar name="f2297"/>',	# &otimes;
	0x2329	=> 'q',						# &lang;
	0x232a	=> 'r',						# &rang;
	0x8482	=> '<userchar name="f8482"/>',	# &trade;
	aacute	=> 0x00e1,
	acute	=> 0x00b4,
	aelig	=> 0x00e6,
	agrave	=> 0x00e0,
	alpha	=> 0x03b1,
	auml	=> 0x00e4,
	Eacute	=> 0x00c9,
	eacute	=> 0x00e9,
	ecirc	=> 0x00ea,
	egrave	=> 0x00e8,
	empty	=> 0x2205,
	hArr	=> 0x21d4,
	iota	=> 0x03b9,
	iuml	=> 0x00ef,
	kappa	=> 0x03ba,
	lambda	=> 0x03bb,
	Lambda	=> 0x039b,
	lang	=> 0x2329,
	larr	=> 0x2190,
	ldquo	=> 0x201c,
	lsquo	=> 0x2018,
	mdash	=> 0x0097,
	micro	=> 0x00b5,
	middot	=> 0x00b7,
	minus	=> 0x2212,
	mu		=> 0x03bc,
	nu		=> 0x03bd,
	omicron	=> 0x03bf,
	oplus	=> 0x2295,
	Oslash	=> 0x00d8,
	otimes	=> 0x2297,
	ouml	=> 0x00f6,
	pi		=> 0x03c0,
	plusmn	=> 0x00b1,
	pound	=> 0x00a3,
	rang	=> 0x232a,
	rarr	=> 0x2192,
	reg		=> 0x00ae,
	rdquo	=> 0x201d,
	rho		=> 0x03c1,
	rsquo	=> 0x2019,
	schwa	=> 0x0259,
	shy		=> 0x00ad,
	sim		=> 0x223c,
	szlig	=> 0x00df,
	times	=> 0x00d7,
	trade	=> 0x8482,
	uuml	=> 0x00fc,
);

my $ENCODING = 'Shift_JIS';
my %JARGCOND = (
	'condition' => 'web',
	'history' => '',
);

sub new
{
	my $proto = shift;
	my $class = ref($proto) || $proto || "XML::LibXML";
	my $xml = $class->SUPER::new();
	$xml->keep_blanks(1);
	$xml->expand_entities(0);
	$xml->load_ext_dtd(0);
	$xml->recover(1);
	eval { $xml->line_numbers(1); };
	bless($xml, $class);
}

sub JargonXml::parse_file($)
{
	my $self = shift;
	my $file = shift;
	my $doc;
	local $_;

	# replace entity references to prevent them from being removed by the
	# parser. attribute nodes should not contain any entity references
	# except built-in ones such as &lt; or &gt;.
	open(IN, $file) || die "Unable to open file: $file ($!)\n";
	while (<IN>) {
		my $line = '';
		s/<\?xml [^>]*\?>/<?xml version="1.0" encoding="$ENCODING"?>/;
		while ('' ne $_) {
			if (s/^&([^ <>&;]+);//) {
				$line .= &entityValue($1);
			} elsif (s/^([^&]+)//) {
				$line .= $1;
			} else {
				$line .= substr($_, 0, 1);
				$_ = substr($_, 1);
			}
		}
		$self->parse_chunk($line);
	}
	close(IN);
	$doc = $self->parse_chunk('', 1);
	&parseConditional($doc);
	$doc;
}

sub entityValue($)
{
	my $name = shift;
	my $code = $name;

	if ($code =~ s/^#(x?)//) {
		$code = $1 ? hex($code) : $code;
	} else {
		$code = $ENTITY{$name} || return "&$name;";
	}
	defined($ENTITY{$code}) ? $ENTITY{$code} : "&#$code;";
}

sub parseConditional($)
{
	my $node = shift;
	my $parent = $node->parentNode();
	my $type = $node->nodeType();

	if ($type == NODE_PI) {
		return if ('if' ne $node->nodeName());

		# PI node always returns empty attributes (which seems against W3C
		# DOM though). use getData instead.
		my ($name, $value) = split(/=/, $node->getData(), 2);
		my ($cond, $curnode, $nextnode);

		$value =~ s/^([\"\'])(.*)\1$/$2/;
		exists($JARGCOND{$name})
			|| warn "Unknown ?if condition: $name=$value\n";
		$cond = $value eq $JARGCOND{$name};
		$curnode = $node;

		while ($curnode) {
			$nextnode = $curnode->nextSibling();
			if (NODE_PI == $curnode->nodeType()) {
				$name = $curnode->nodeName();
				last if ($name eq 'fi');
				$cond = !$cond if ($name eq 'else');
			} elsif (!$cond) {
				$parent->removeChild($curnode);
			}
			$curnode = $nextnode;
		}

	} elsif ($type == NODE_COMMENT) {
		$node->parentNode()->removeChild($node);

	} elsif ($type == NODE_ENTITY_REF) {
		# nop. walking through the children of an entity reference would
		# resulting in an infinit loop.

	} else {
		my @child = $node->childNodes();
		while (@child) {
			&parseConditional($child[0]);
			shift(@child);
		}
	}
}

1;
